# charlemagne

## a physics file creator

`charlemagne` can read JSON files and serialize them as binary physics
files.  It provides a number of options that can facilitate the
management of large projects involving multiple physics files.  Once
you begin using JSON as your base format, you should be able to avoid
editing the binary files, period, if you so desire.

### getting started

I don't recommend creating the JSON by hand.  You'll probably want to
generate a template from some base physics file you've already cooked
up.  Or you can start from the "Standard" physics file from Infinity.

    go build
    ./charlemagne -i Standard.phyA -o Standard.json # creates monolithic JSON
    # edit JSON as desired
    ./charlemagne -i Standard.json -j -o Standard.phyA # create binary from JSON

Note the `-j` flag in the last invocation: this tells `charlemagne`
that its input is a JSON file and not (as is the default) a physics
binary.  You can unambiguously mark the input as binary using the
`-b` flag, in case you're like me and want orthogonal commands where
sensible:

    ./charlemagne -i Standard.phyA -b -o Standard.json # creates monolithic JSON

Unlike with binary-to-JSON conversion, you *must* specify an output
file when converting JSON to binary.  Part of this limitation is
technical in nature, but part is also for your sanity in case you
forget to redirect the output on your terminal.  If demand is great
enough, I'll consider adding binary output to stdout.

### assemblages

While it's nice to be able to store isolated physics JSON
monolithically, there are a number of instances in which you might
wish for a slightly more modular approach.  For example, you might
have slightly (or vastly) different physics files to be merged into a
number of levels, and want these files to share a smaller set of
backing data.  More concretely, let's say the Tick monster is the same
across all levels, but the Drone's physics/graphics need to change
depending on the level.  If you have 50 levels, any change to the Tick
would require updating 50 separate files; then you'd have to remember
which variant of the Drone you're updating and what levels use that
variant.

You can create "assemblages" for this purpose.  These rely on
entity-level JSON files as seen with the `-e` flag (more detail in the
section called `a physics file transformer`.  An "assemblage" is a
top-level JSON file that lists the individual source files that
constitute a complete physics file.  To get started:

    ./charlemagne -i Stuff.phyA -e -a -o physics # outputs to directory physics
    # last line of output will list the assemblage file: physics/physics.json

You'll end up with a directory `physics` that contains subdirectories
for entity type, each of which contains an individual JSON file per
entity.  The assemblage file `physics.json` lists all of the paths to
these individual JSON files.  Unless the paths are absolute
(i.e. begin with a `/` character on real computers or a `\` character
on fake ones), they represent the location of a file relative to the
assemblage file.

It also preserves the filename of the overlay you used, if any, when
generating the files, and reversing the process will attempt to use
this same overlay.  See the `name overlays` section for more
information about overlays.

Once you've generated the first assemblage file and its consituents,
you can do this, following the above example:

    mkdir common
    mv physics/* common/
    mv common physics/
    mv physics/common/physics.json physics/common.json
    
    # below: we prepend all paths with "common/"
    sed -i.bak -e 's/\"\([a-z]\+\/\)/"common\/\1/' physics/common.json
    
    mkdir -p physics/level01/monsters
    cp physics/common/monsters/drone-minor.json physics/level01/monsters/drone-minor.json
    
    # edit the new drone-minor.json file as needed
    
    cp physics/common.json physics/level01.json
    
    # edit level01.json so the drone-minor.json monsters entry is:
    # physics/level01/monsters/drone-minor.json
    
    ./charlemagne -i physics/common.json -j -o Common.phyA
    ./charlemagne -i physics/level01.json -j -o Level01.phyA

In this example, we use the initial JSON as the "common" set of
physics.  Then we create a second assemblage file for a specific
level, pull out the one thing that needs to be different, and cause
the level-specific file to pull that in.  You can organize things
however you like, and obviously a large project would have many
different files.  If your levels have more general "flavors" or
"environments" and the physics follow suit, you of course don't need
to have an assemblage file per level:

    ./charlemagne -i physics/pfhor-prime.json -j -o 00\ Touchdown.phyA
    ./charlemagne -i physics/boomer.json -j -o 01\ Rescued.phyA
    ./charlemagne -i physics/pfhor-prime.json -j -o 02\ Back\ At\ It.phyA

As you can see, any updates to the specific entities pointed out in an
assemblage file will then be propagated as needed.  From here, you can
write scripts or even a makefile that will generate the final physics
binaries and e.g. put them into the proper merge directory structure
for a tool like `atquem` to use.

This is all a lot of work, granted, and you might already be up the
creek by having a ton of physics files with the common stuff manually
distributed between them.  Hopefully, if it's too much effort to
identify all the commonalities, you can at least use monolithic JSON
files and use tools to update them in batches.  Or you can use
`charlemagne` to generate JSON for all of your physics files, compare
the output, and see where commonalities lie (and probably also see
some places you forgot about).  If you're doing anything more than
maintaining a project, I'm willing to bet that going through the
effort of using assemblages will be worth it in the long run.

The sharp-eyed among my readers will have noticed that the
"assemblage" JSON is really tacked onto the existing monolithic JSON
schema.  You can in fact combine the two, but assemblage will always
override whatever is formally defined - even if the assemblage is
empty.  I recommend not mixing these in the same file.

### final binary-generation notes

The names of individual files (or even the `name` field on each
entity) don't determine what "slot" those entities map to in the final
file; instead, the position in the array of entities/sources indicates
the slot.  In other words, using an overlay won't matter in terms of
figuring out which entity JSON file associates to which slot; however,
the overlay does apply to the contents of these files.  For example,
you might name the first projectile file in the sources list
`fists-of-fury.json`, but if the overlay calls that first projectile
`fists`, anything referring to that projectile in your other files
will need to use the string `fists` instead of `fists-of-fury`.  Or if
you use a monolithic file and change the first projectile entity's
`name` to `fists-of-fury`, this does not affect mapping either.  If
this is all confusing, don't worry; hopefully things will mostly work
as you expect.  If it doesn't, you'll need to use (or update) your
overlay file.

I don't believe there are any hard-coded limits to the number of
entities that can be in a given section's list, at least on
`charlemagne`'s end.  You should not generate physics files with fewer
entities than in the standard Infinity physics file.  I can't say what
will happen if you add additional entities (e.g. define 500 monsters),
but I doubt it will work in the engine the way you'd like.  If you
want to try this, you will first need to create a custom overlay and
then add entries for each new index/name.  My advice is to keep it all
in version control, just in case.

## a physics file transformer

As noted above, `charlemagne` can also read a physics file and
serialize it as JSON.  The intention is to facilitate change tracking
against files that are otherwise less-than-diffable for human beings.
To this end, it can output a single monolithic JSON file representing
the entire structure of the physics data; as a set of five files, one
per data section, or as a set of files by individual entry.

    go build
    ./charlemagne [-s | -e <-o outdir>] [-o outfile] [ -L overlay.[csv | json] ] [ -r ] <[ -i input.phyA] | input.phyA>
    ./charlemagne -defaults

- Invoke with `-s` or `-e` to enable splitting by section or entry
    - must invoke also with `-o <outdir>` to specify the output
      directory for the section files
    - a "section" is one of `monsters`, `effects`, `projectiles`,
      `constants`, or `weapons`; the `<section>.json` file will
      contain a list of all entries for the data in that section.
    - an "entry" is a single datum.  For example, the "Fist" weapon
      definition in the `weapons` section.  If splitting by entry,
      you'll end up with a set of subdirectories in `<outdir>`
      named for each section, and within each subdirectory, a
      `<entry>.json` file named after a slightly-transformed
      version of the entry's name.
- If not splitting, you can omit the `-o outfile` flag in order to
  output to standard out.  With or without the output file
  specified, you'll end up with JSON that contains the physics data
  in its entirety.  An output file name of `-` is the same as
  standard output.
- You may specify the input file either with `-i input.phyA` or just
  by giving the input file at the end of the list of arguments.
- You may pipe into standard input rather than choose an input file
  name.  You may also use `-i -` to do this.
- You may use a name-mapping overlay with the `-L` flag to output
  various `Name` fields (and file names) based on your own names for
  various entities.  See `name overlays` for details.
  Must be a file whose name ends in `.json` or `.csv` and is actually
  a JSON or CSV file, respectively.
- You may dump out mappings of the default names for various
  entities using the `-defaults` flag.  Note that `charlemagne` will
  do nothing else; also note that the output filenames are fixles
  to ./default-names.json and ./default-names.csv
- Invoke with `-r` to attempt to recover the file from light mangling
  from an editor such as PhysicsEditorOne.  Very light mangling.
  
## name overlays

In case you don't like the names I've chosen for various entities in
the physics files, or if you're working on a scenario (dare I say
total conversion) with which it would make more sense to use your own
names for things, you may easily do so.  Just keep in mind that any
changes to a name overlay will bubble down into the program's output.
This might cause (diff) changes to files that otherwise didn't change.
If you start using an overlay and are checking things into version
control, I recommend doing this as its own commit.

For example, I've preserved ShapeFusion's strings in
[JSON](./overlays/shapefusion-names.json) and
[CSV](./overlays/shapefusion-names.csv) form.  If you want to dump out
a sample, invoke `./charlemagne -defaults` and edit to your heart's
content.

Note that you can selectively remap: if a file has a single remapping,
say calling the monster that we usually know as the "player" or
"marine" by some other name, charlemagne will preserve the defaults.
Your overlay file need not be comprehensive.

Maybe at some point I'll allow actual overlays, i.e. layers of
remappings.  Not now.

## helper scripts

A number of helper scripts are in the `scripts` directory.  Details in
the corresponding [README](./scripts/README.md).  Note that you don't
need these scripts if you are modifying JSON files and generating the
binaries; these are mainly useful to you if you're dealing entirely
with binary physics files.

- `physics_file_diff`: compare two files, two `git` commits, or a
   mixture.
- `physics_file_history`: generate `git` commit history for the JSON
   form of some physics file.
- `post-commit.sample`: a sample `git` post-commit hook that
   automatically generates JSON for any `.phyA` files and amends the
   commit to include these.

## TODO

- Better unit tests.
- Layered overlays.

## credits

Without Tito Dal Canton, Etienne Sampson, et al's work on ShapeFusion,
I would never have been willing to slog through the Aleph One source
code to figure out the file format.  Nor would I have bothered with
naming anything; their names for various entities were extremely
helpful not only in the output you see now but also in figuring out
just how badly I had messed up reading some part of a file.

## license (two-clause BSD)

Copyright (c) 2021, Jonathan G. Irons

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
