# Hacky Project Management Scripts

This set of scripts sets up a consolidated set of entity-level physics
JSON files that you can then use to manage a large scenario.  What
does this mean to you?  Once you've run these scripts, you'll have a
single tree of physics JSON organized by "section" (constants,
monsters, projectiles, etc.) and then by entity names and variants.

For example, if your scenario uses 10 distinct physics files, chances
are you have some identical stuff common to all files, some stuff that
is unique to a given file, and a range of commonality in between.
Your scenario might have the same walking physics for all levels, have
differing juggernaut parameters for each level, and an even split of
grenade behavior between two sets of parameters.  The scripts will not
only split all these out, but it will detect all common definitions
and set up assemblage files that allow you to re-use a single JSON
file for a distinct definition.  This means you don't have to edit
the walking physics in ten places every time you make a change.

You'll need `bash`, `charlemagne` (in your `PATH`), and `perl`.  Also
some other common utilities like `find`.

## Assumptions

* The current directory (i.e. the one with these scripts) also has
  each `phyA` binary file you wish to manage.
* You don't have one called `entities.phyA`.
* You don't need name overlays.  When testing this with Eternal,
  I found a few instances of sound indexes that went beyond those
  in the default list.  In Eternal's case, I've been nice enough
  to include overlays.  If you need to shore up mappings for your
  own scenario, run `charlemagne -defaults`, copy either the
  resulting JSON or CSV file, edit it, and then modify `get-sums.sh`
  and `generate.sh` so that the `charlemagne` invocations also
  have the parameter `-L <your overlay file>`.  You'll get hard
  errors if you need mappings.  Should be pretty obvious.

## `do-everything.sh`

This script runs all the following scripts in the order listed below.

## `get-sums.sh`

From the list of all `.phyA` files in the current directory, uses
charlemagne to generate assemblages such that from file `x.phyA` you
will have a directory called `x` with the full entity-level structure
beneath it.  From the complete set of all JSON across all physics
files, get the sha256 sum of each file and output it to standard out.

## `organize.pl`

Assuming that standard input is the sha256 output from `get-sums.sh`
(i.e. checksum + whitespace + filename), uses this list of sums to
find all variations of each entity.  Copies each variation into a new
hierarchy under an `entities` subdirectory, with a variant ID appended
to each filename, e.g.:

* `./entities/monsters/trooper-major_aa.json`
* `./entities/monsters/trooper-major_ab.json`

One assemblage JSON file will be created directly under `entities` for
each physics file that was split apart by charlemagne in the first
step.  The assemblage file will point to the entities/variants
relevant to the original physics.

Once everything is organized, you can use methods like the following
to get information on which variant is used in which file:

`find ./entities -mindepth 1 -maxdepth 1 -type f -name "*.json" -exec grep -l "trooper-major__af.json" {} \+`

If you want more specific names for these entities, you can rename the
variant JSON file (e.g. change the file name `trooper-major__aa.json`
to `super-trooper.json`) and then replace all occurrences in the
assemblages:

`find ./entities -mindepth 1 -maxdepth 1 -type f -name "*.json" -exec sed -i.bak -e 's/trooper-major__aa.json/super-trooper.json/'`

As always, I highly recommend using version control.

## `generate.sh`

Takes the assemblage files in `entities` and creates a physics file of
the same name, also in `entities`.  Once you've organized the first
time, you can just use this script to conveniently generate all your
binary files, then ship them to whatever place is useful (probably
map merge directories).
