#!/usr/bin/perl -w

use strict;
use utf8;

my %sums_inputs;
my %inputs_sums;
my %sums_bases;
my %sums_counts;
my %sums_outputs;
my @counts;

my @alpha = (
    'a', 'b', 'c', 'd', 'e', 'f', 'g', 'h',
    'i', 'j', 'k', 'l', 'm', 'n', 'o', 'p',
    'q', 'r', 's', 't', 'u', 'v', 'w', 'x',
    'y', 'z'
    );

my @subdirs = (
    "constants", "effects", "monsters",
    "projectiles", "weapons"
    );

my $TOP_DIR = "entities";

my $alpha_len = @alpha;

sub build_lookups
{
    my $sum_input;
    my $n;
    my $sum;
    my $input;
    my $base;
    my $count;
    my @pair;
    
    while (<>) {
	m|\..+/.+/.+.json| or next;
	chomp;
	$sum_input = $_;
	@pair = split /\s+/, $sum_input;
	if (@pair < 2) {
	    die "input line does not split to two items: ${sum_input}";
	}

	$input = $pair[1];
	$n = 2;
	
	$sum = $pair[0];

	while ($n < @pair) {
	    $input .= ' ' .$pair[$n];
	    $n++;
	}
	
	$base = $input;
	$base =~ s|.*/([^/]+/.*?.json)|$1|;

	$inputs_sums{$input} = $sum;
	
	my $table;
	
	if (exists($sums_inputs{$sum})) {
	    $table = $sums_inputs{$sum};
	} else {
	    my @tmp = ();
	    $table = \@tmp;
	    $sums_inputs{$sum} = $table;
	}

	$count = @{ $table };
	
	push @{ $table }, $input;
	
	$sums_counts{$sum} = $count;
	$sums_bases{$sum} = $base;
    }

    my $len;

    while (my ($key, $value) = each(%sums_counts)) {
	$len = @counts;
	while ($len < $value) {
	    push @counts, [];
	    ++$len;
	}

	push @{ $counts[$value] }, $key;
    }

    $len = @counts;

    my %bases_counter;
    
    while (my ($key, $value) = each(%sums_bases)) {
	my $rank;
	if (exists($bases_counter{$value})) {
	    $rank = $bases_counter{$value};
	} else {
	    $rank = 0;
	}

	my $i = $len;
	while ($i > 0) {
	    --$i;
	    my @sum_counts = @{ $counts[$i] };
	    foreach $sum (@sum_counts) {
		if ($sum eq $key && !exists($sums_outputs{$sum})) {
		    my $suffix = &int_to_alpha($rank, 2);
		    my $whole = $value;
		    $whole =~ s|\.json|__$suffix.json|;
		    $sums_outputs{$sum} = "${whole}";
		}
	    }
	}
	
	$bases_counter{$value} = $rank + 1;
    }
}

sub copy_variants
{
    -d $TOP_DIR || mkdir $TOP_DIR || die "Could not create dir ${TOP_DIR}";
    foreach my $subdir (@subdirs) {
	$subdir = "${TOP_DIR}/${subdir}";
	-d $subdir || mkdir $subdir || die "Could not create dir ${subdir}";
    }
    my $x = 0;
    while (my ($sum, $inputs) = each(%sums_inputs)) {
	# we only need the first input because the files with
	# the same sum have the same contents.
	my $input = @{ $inputs }[0];
	my $output = $sums_outputs{$sum};
	
	$output = "${TOP_DIR}/${output}";
	open(my $of, '>', $output) || die "Could not open output file ${output}";
	open(my $if, '<', $input) || die "Could not open input file ${input}";

	while (<$if>) {
	    print $of $_;
	}
	
	close $if;
	close $of;
    }
}

sub rewrite_assemblages
{
    # we're just lazy and get the json files in top-level
    # dirs, except for those in TOP_DIR

    my @listings = glob("*");

    foreach my $listing (@listings) {
	if ($listing eq $TOP_DIR || "./${listing}" eq $TOP_DIR) {
	    next;
	}

	if (-d $listing) {
	    my $assemblage = "${listing}/${listing}.json";
	    my $rewritten = $assemblage;
	    $rewritten =~ s|$listing/|$TOP_DIR/|;
	    
	    open(my $if, '<', $assemblage) || die "Could not open input assemblage ${assemblage}";
	    open(my $of, '>', $rewritten) || die "Could not open output assemblage ${rewritten}";
	    
	    while (<$if>) {
		chomp;
		my $line = $_;
		my $text = $line;
		my $wow_commas = 0;
		$text =~ s|^\s+"([^/]+/[^/]+.json)",?.*$|$1|;
		my $input = $text;
		$input = "./${listing}/${text}";
		if (exists($inputs_sums{$input})) {
		    $wow_commas = m/"$text",/;
		    my $sum = $inputs_sums{$input};
		    my $output = $sums_outputs{$sum};
		    $text =~ s|$text|$output|;
		    if ($wow_commas) {
			print $of "   \"${text}\",\n";
		    } else {
			print $of "   \"${text}\"\n";
		    }
		} else {
		    print $of "${line}\n";
		}
	    }
		
	    close $if;
	    close $of;
	}
    }
}

sub int_to_alpha
{
    my $n = shift;
    my $len = shift || 1;
    my $str = "";
    my $total = 0;
    my $m;

    while ($n || $total < $len) {
	$m = $n % $alpha_len;
	$str = "${alpha[$m]}${str}";
	$n = int(int($n) / int($alpha_len));
	++$total;
    }

    return $str;
}

sub main
{
    &build_lookups;
    &copy_variants;
    &rewrite_assemblages;
}


&main;
