#!/usr/bin/env sh

# charlemagne needs to be in your PATH, or use a local copy adjacent
# to this script.

if [[ "$#" -lt 2 ]];
then
    echo "usage: ${0} <patch file> file1.phyA [ file2.phyA ... ]"
    exit 1
fi

charlie=$(which charlemagne)

[[ -e "./charlemagne" ]] && charlie="./charlemagne"

conversion_patch="$1"; shift;
tempdir="./.tmp"
infdir="./infinity-entities"

for to_patch in "$@";
do
    (echo "${to_patch}" | grep --silent ".phyA$") || continue

    echo "${to_patch}:"

    rm -rf "${tempdir}" || exit 1;

    echo "  generating JSON entities"
    "${charlie}" -b -i "${to_patch}" -e -o "${tempdir}" 1>/dev/null || exit 1
    
    echo "  patching JSON"
    patch --silent "${tempdir}/.tmp.json" "${conversion_patch}" || exit 1

    echo "  copying new entities"

    rsync --quiet -a "${infdir}/" "${tempdir}" || exit 1
    
    echo "  creating backup ${to_patch}.orig"
    cp "${to_patch}" "${to_patch}.orig" || exit 1
    
    echo "  generating binary file"
    "${charlie}" -j -i "${tempdir}/.tmp.json" -e -o "${to_patch}" || exit 1
    echo "  ...OK"
done

rm -rf "${tempdir}" || exit 1
