# charlemagne bash scripts

## `physics_file_diff`

This script (which we'll call `pfd` from now on - feel free to alias
or link or rename it as such for your own use) uses `charlemagne` to
convert two physics files into basic, non-split JSON output and runs
the result through `diff -u`.  The physics files may be real files in
situ on the file system, or may be linked to commits.  You may also
feed the contents one file into standard input.

Specify one commit with a single `-c <commit>` argument, or two
commits with two such arguments. The file(s) to which the commit(s)
will apply are context-sensitive.  A file that does not have a commit
applied to it will use the current on-FS version of the file.  If you
don't care about commits, you don't need to worry about `git` at all.

Examples:

    ./pfd A.phyA B.phyA                               # compare two files
                                                      # from the filesystem
    
    cat A.phyA | ./pfd B.phyA                         # compare standard input
                                                      # with a file on the
                                                      # filesystem.
    
    ./pfd <commit-for-A> A.phyA                       # compare a commit with
                                                      # the on-filesystem
                                                      # version of the file
    
    ./pfd <commit-1> <commit-2> A.phyA                # compare two commits
                                                      # for the same file
    
    ./pfd <commit-for-A> <commit-for-B> A.phyA B.phyA # compare two commits
                                                      # for different files
    
    cat A.phyA | ./pfd <commit-for-B> B.phyA          # compare standard input
                                                      # to a commit of a file
    
    cat A.phyA | ./pfd <commit>                       # invalid

You can change the amount of context in the diff.  Run with the `-h`
flag to see all options.

## `post-commit.sample`

This is a sample `git` post-commit hook that examines the commit that
just happened for files with the extension `.phyA`.  It generates JSON
output for any files it finds and amends the commit so that these files
are also part of it, tying the binary and human-readable changes into
the same commit.

You'll either need to `export CHARLEMAGNE_PATH=/path/to/charlemagne`
in your `.bashrc` (or however you system sets environment variables),
or you'll need to put `charlemagne` in a place your `PATH` environment
variable can see it.  Or set `CHARLEMAGNE_PATH` straight in the
script.  This might take some trial and error depending on how you run
`git` commands and what `git` can see of your environment and
filesystem.  Then:

    cp post-commit.sample /path/to/your/repo/.git/post-commit

The sample script splits each file at the "entry" level and places
output in a directory adjacent to that file called
`readable-<filename>`.  It is easy enough to change this name or to
split at the "section" level (or not at all).  You'll need to know a
little bash if you want the script to behave differently beyond that -
if, for example, you have a parallel location in the repo where the
JSON should reside.

With things split at the entity level, you will be able to examine a
commit (or its diff) using the usual tools and visualizations and get
a pretty clear picture of what things changed in the physics file just
by seeing which JSON files were modified.  Of course, looking at the
file diffs will give you even more information.  Also, you can modify
this script to use your own overlay if you want strings and
entity-level files to follow what your scenario calls them.

Important note.  While this seems to work well in my limited testing,
the mechanism by which this script prevents an infinite loop of
amendments seems pretty delicate to me.  You have been warned.  If a
minor commit seems to hang, chances are it is indeed hanging.  I can't
vouch for the exact state of your commit if you send SIGINT
(control-C) to the process, etc., but it's better than having an
infinite loop.  Again, this didn't happen when I tested.  Just
something to look out for.

## `physics_file_history`

Assuming you already have extensive commit history for physics files
before using the post-commit hook, you might need to build a parallel
history some other way.  This script (which we'll call `pfh` from now
on - feel free to alias or link or rename it for as such for you own
use) looks at the history for a physics file on a git repository.  It
creates (or re-uses) a branch, generating `charlemagne` output for
each commit against the file and making its own commit with the
output, adding some information to the commit message along the way.
The branch can then be merged to another, if you don't mind a bunch of
these "history" commits, or can be kept on its own for an easy-to-read
change history.

This can supplement any other scripts that might come out of the
charlemagne project, such as a planned commit hook that will run
`charlemagne` against any committed files and put them somewhere
useful.  `pfh` will let you retroactively record changes, while such a
hook will let you keep track from now on.

You can run `pfh` to get a sense of its limited options.  Perhaps the
best documentation is my final set of tests against some prominent
mararthon projects on github.  Note that I have set my
`CHARLEMAGNE_PATH` environment variable to the location of the
`charlemagne` executable.  This value could also be present on my
`PATH` (i.e. within reach of `which`), or you can supply the script
with the `-x` parameter to point to it.

    git clone https://github.com/Pfhorrest/Eternal.git
    cd Eternal
    pfh -b test_physics_hist -m "readable physics" -i Eternal-Physics.phyA -o readable_physics -e
    first_hash=$(git --no-pager log --reverse --pretty=format:"%h %s" | grep -- '-> JSON' | head -c 7)
    git diff "${first_hash}" HEAD
    git diff HEAD~1 HEAD

    cd Physics-model-templates
    pfh -b test_physics_hist -m "readable physics" -e -op readable_ *.phyA

Note the use of the `-op` flag in combination with a set of files for
the second example.  Unfortunately, `pfh` doesn't chase files that
have been renamed or moved.  I doubt that will happen, but who knows.

## TODO

u tell me

## license (two-clause BSD)

Copyright (c) 2021, Jonathan G. Irons

Redistribution and use in source and binary forms, with or
without modification, are permitted provided that the following
conditions are met:

1. Redistributions of source code must retain the above copyright
notice, this list of conditions and the following disclaimer.

2. Redistributions in binary form must reproduce the above copyright
notice, this list of conditions and the following disclaimer in the
documentation and/or other materials provided with the distribution.

THIS SOFTWARE IS PROVIDED BY THE COPYRIGHT HOLDERS AND CONTRIBUTORS
"AS IS" AND ANY EXPRESS OR IMPLIED WARRANTIES, INCLUDING, BUT NOT
LIMITED TO, THE IMPLIED WARRANTIES OF MERCHANTABILITY AND FITNESS FOR
A PARTICULAR PURPOSE ARE DISCLAIMED. IN NO EVENT SHALL THE COPYRIGHT
OWNER OR CONTRIBUTORS BE LIABLE FOR ANY DIRECT, INDIRECT, INCIDENTAL,
SPECIAL, EXEMPLARY, OR CONSEQUENTIAL DAMAGES (INCLUDING, BUT NOT
LIMITED TO, PROCUREMENT OF SUBSTITUTE GOODS OR SERVICES; LOSS OF USE,
DATA, OR PROFITS; OR BUSINESS INTERRUPTION) HOWEVER CAUSED AND ON ANY
THEORY OF LIABILITY, WHETHER IN CONTRACT, STRICT LIABILITY, OR TORT
(INCLUDING NEGLIGENCE OR OTHERWISE) ARISING IN ANY WAY OUT OF THE USE
OF THIS SOFTWARE, EVEN IF ADVISED OF THE POSSIBILITY OF SUCH DAMAGE.
